using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using System.Threading;
using static MeshVR.PresetManager;

namespace JayJayWon
{

    public class ExternalButtonProvider
    {
        private static List<ExternalButtonProvider> providers= new List<ExternalButtonProvider>();
        private static Dictionary<MVRScript, ExternalButtonProvider> providersByMVRScript = new Dictionary<MVRScript, ExternalButtonProvider>();

        private static List<ExternalButtonProvider> activeProviders { get { return providers.Where(x => x.buttonNameList.Count > 0).ToList(); } }
        public static ExternalButtonProvider currentProvider { get; private set; } = null;

        public static bool available { get { return activeProviders.Count > 0; } }

        public static void RegisterProvider(MVRScript script, Dictionary<string, object> buttonListData)
        {
            UnregisterProvider(script);

            if (buttonListData.ContainsKey("ButtonNameList") && buttonListData.ContainsKey("ButtonPressCallbackJSS"))
            {
                var provider = new ExternalButtonProvider(script, buttonListData);
                providers.Add(provider);
                providersByMVRScript.Add(script, provider);

                if ((currentProvider==null || !activeProviders.Contains(currentProvider)) && activeProviders.Contains(provider)) currentProvider =provider;           
            }

            if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.externalButtonProviders) GameControlUI.Update();
        }

        public static bool UnregisterProvider(MVRScript script)
        {
            if (providersByMVRScript.ContainsKey(script))
            {
                var provider = providersByMVRScript[script];
                providers.Remove(provider);
                providersByMVRScript.Remove(script);

                if (currentProvider == provider)
                {
                    if (activeProviders.Count>0) currentProvider= activeProviders.First();
                    else currentProvider = null;
                }
                return true;
                
            }
            return false;
        }

        public static bool leftProvidersAvailable { get
            {
                if (activeProviders.IndexOf(currentProvider)>0) return true;
                return false;
            } }

        public static bool rightProvidersAvailable
        {
            get
            {
                if (activeProviders.IndexOf(currentProvider) < providers.Count-1 ) return true;
                return false;
            }
        }

        public static void SelectFirstProvider()
        {
            if (activeProviders.Count>0) currentProvider = activeProviders.First();
            else currentProvider=null;
        }

        public static void SelectLastProvider()
        {
            if (activeProviders.Count > 0) currentProvider = activeProviders.Last();
            else currentProvider = null;
        }

        public static void SelectPreviousProvider()
        {
            if (activeProviders.Count > 1 && currentProvider!= activeProviders.First()) currentProvider = activeProviders[activeProviders.IndexOf(currentProvider)-1];
            else currentProvider = activeProviders.FirstOrDefault();
        }

        public static void SelectNextProvider()
        {
            if (activeProviders.Count > 1 && currentProvider != activeProviders.Last()) currentProvider = activeProviders[activeProviders.IndexOf(currentProvider) + 1];
            else currentProvider = activeProviders.LastOrDefault();
        }

        public static List<string> currentOptionNames { get
            {
                if (currentProvider == null) return new List<string>();
                return currentProvider.optionNames;
            } }

        

        private JSONStorableString buttonPressCallbackJSS;
        private JSONStorableAction buttonRefreshRequestCallbackJSA;
        private List<string> buttonNameList;
        private Dictionary<string, Color> buttonColorsByName;
        private Dictionary<string, Color> buttonTextColorsByName;
        private MVRScript script;
        public string externalProviderName { get; private set; }
        public ExternalButtonProvider(MVRScript script, Dictionary<string, object> buttonListData)
        {
            
            this.script = script;
            ProcessButtonListData(buttonListData);
        }

        public void SetButtonColors(UIDynamicButton button, string optionName)
        {
            if (buttonColorsByName != null && buttonColorsByName.ContainsKey(optionName))
            {
                button.buttonColor = buttonColorsByName[optionName];
            }

            if (buttonTextColorsByName != null && buttonTextColorsByName.ContainsKey(optionName))
            {
                button.textColor = buttonTextColorsByName[optionName];
            }
        }


        private void ProcessButtonListData(Dictionary<string, object> buttonListData)
        {
            if (buttonListData.ContainsKey("ButtonPressCallbackJSS")) buttonPressCallbackJSS = buttonListData["ButtonPressCallbackJSS"] as JSONStorableString;
            else buttonPressCallbackJSS = null;      

            if (buttonListData.ContainsKey("ForceButtonListDataRefresh")) buttonRefreshRequestCallbackJSA = buttonListData["ForceButtonListDataRefresh"] as JSONStorableAction;
            else buttonRefreshRequestCallbackJSA = null;

            if (buttonListData.ContainsKey("ButtonNameList")) buttonNameList = buttonListData["ButtonNameList"] as List<string>;
            else buttonNameList = null;

            if (buttonListData.ContainsKey("ButtonListProviderName")) externalProviderName = buttonListData["ButtonListProviderName"] as string;
            else externalProviderName = script.containingAtom.uid +": "+script.storeId;

            if (buttonListData.ContainsKey("ButtonColorsByName")) buttonColorsByName = buttonListData["ButtonColorsByName"] as Dictionary<string, Color>;
            else buttonColorsByName = null;
            if (buttonListData.ContainsKey("ButtonTextColorsByName")) buttonTextColorsByName = buttonListData["ButtonTextColorsByName"] as Dictionary<string, Color>;
            else buttonTextColorsByName = null;

            if (buttonRefreshRequestCallbackJSA!=null) buttonRefreshRequestCallbackJSA.actionCallback = RefreshButtonListData;

            
        }

        private void RefreshButtonListData()
        {
            var buttonListData = new Dictionary<string, object>();

            bool refresh = currentProvider == this;

            if (script!= null)
            {
                script.SendMessage("OnUIAButtonListDataRequested", buttonListData, SendMessageOptions.DontRequireReceiver);
                ProcessButtonListData(buttonListData);

                if (activeProviders.Count == 0) currentProvider = null;
                else if (!activeProviders.Contains(currentProvider))
                {
                    currentProvider = activeProviders[0];
                    refresh = true;
                }
            }
            

            if (refresh && GameControlUI.gameControlDisplayMode == GameControlDisplayModes.externalButtonProviders)
            {
                GridsDisplay.uiExternalOptionSelector.DestroyUI();
                GridsDisplay.uiExternalOptionSelector.CreateUI();
            }

            if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.externalButtonProviders && currentProvider==null)
            {
                GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard;
                GameControlUI.RefreshWristUIButtonGrid();
            }
            
        }

        public List<string> optionNames { get {
                return new List<string> (buttonNameList); } }

        public void OptionPressed (string optionName)
        {
            if (buttonPressCallbackJSS!=null) buttonPressCallbackJSS.val = optionName;
        }

    }
}
