using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace JayJayWon
{

    public static class VAMStorablesCatalog
    {
        private static Dictionary<string, VAMStorablesAtomType> atomTypes =null;

        private static List<VAMStorablesStorable> allAtomStorables=null;

        public static void CreateCatalog(JSONNode atomCatalog, JSONNode personCatalog)
        {
            atomTypes = new Dictionary<string, VAMStorablesAtomType>();
            

            foreach (JSONClass atomTypeJC in atomCatalog["atomTypes"] as JSONArray)
            {
                CreateAtomType(atomTypeJC);
            }

            JSONClass personAtomTypeJC = personCatalog["basePerson"].AsObject;
            personAtomTypeJC["atomTypeID"] = "Person";
            CreateAtomType(personAtomTypeJC);

            var personAtomType = atomTypes["Person"];

            foreach (JSONClass personVariantJC in personCatalog["personVariants"] as JSONArray)
            {
                personAtomType.ProcessPersonVariant(personVariantJC);
            }
        }

        private static void CreateAtomType (JSONClass atomTypeJC)
        {
            var atomType = new VAMStorablesAtomType(atomTypeJC);
            atomTypes.Add(atomType.atomTypeID, atomType);
        }

        public static List<string> GetStorableNames(string atomType)
        {
            return atomTypes[atomType].GetStorableNames();
        }

        public static List<string> GetParamNames(string atomType, string receiverName)
        {
            return atomTypes[atomType].GetParamNames(receiverName);
        }

        public static JSONStorable.Type GetJSONStorableType (string atomType,string receiverName, string targetName)
        {
            return atomTypes[atomType].GetJSONStorableType(receiverName, targetName);
        }


        public static void ResetJSONParam(string atomType, string receiverName, string targetName, JSONStorableParam param)
        {
            atomTypes[atomType].ResetJSONParam(receiverName, targetName, param);
        }


    }

    public class VAMStorablesAtomType
    {
        public string atomTypeID { get; private set; }

        private Dictionary<string, VAMStorablesStorable> storablesByName = new Dictionary<string, VAMStorablesStorable>();

        public List<VAMStorablesStorable> storables { get
            {
                return new List<VAMStorablesStorable>(storablesByName.Values);
            } }

        public VAMStorablesAtomType(JSONClass atomTypeJC)
        {
            atomTypeID = atomTypeJC["atomTypeID"];

            foreach(JSONClass storableJC in atomTypeJC["storables"].AsArray)
            {
                var storable = new VAMStorablesStorable(storableJC);
                storablesByName.Add(storable.storableID, storable);
            }
          
        }

        public void ProcessPersonVariant(JSONClass personVariantJC)
        {
            foreach (JSONClass storableJC in personVariantJC["storables"].AsArray)
            {
                string storableID = storableJC["storableID"];
                if (storablesByName.ContainsKey(storableID)) storablesByName[storableID].ProcessStorableVariations(storableJC);
                else
                {
                    var storable = new VAMStorablesStorable(storableJC);
                    storablesByName.Add(storable.storableID, storable);
                }
            }
        }

        public List<string> GetStorableNames()
        {
            return new List<string>(storablesByName.Values.Where(x=>x.paramCount>0).Select(x=>x.storableID));
        }

        public List<string> GetParamNames(string storableName)
        {
            return storablesByName[storableName].GetParamNames();
        }

        public JSONStorable.Type GetJSONStorableType(string storableName, string targetName)
        {
            return storablesByName[storableName].GetJSONStorableType(targetName);
        }

        public void ResetJSONParam( string storableName, string targetName, JSONStorableParam param)
        {
            storablesByName[storableName].ResetJSONParam( targetName, param);
        }
    }

    public class VAMStorablesStorable
    {
        string[] paramTypes = {"Float","Bool","Color","Vector3","String","Url","StringChooser","StringChooserAction","Action","PresetFilePathAction","AudioClipAction" };

        public string storableID { get; private set; }

        public int paramCount { get
            {
                return paramsByType.Count;
            } }

        private Dictionary <string, VAMStorablesParam> paramsByID = new Dictionary<string, VAMStorablesParam>();
        private Dictionary<string, HashSet<VAMStorablesParam>> paramsByType = new Dictionary<string, HashSet<VAMStorablesParam>>();

        public VAMStorablesStorable(JSONClass storableJC)
        {
            storableID = storableJC["storableID"];

            var paramTypesJC = storableJC["paramTypes"].AsObject;

            foreach (string paramType in paramTypes)
            {                
                if (paramTypesJC[paramType] != null)
                {
                    foreach (JSONClass paramJC in paramTypesJC[paramType].AsArray)
                    {
                        AddParam(paramJC,paramType);
                    }
                }
            }
            
        }

        private void AddParam(JSONClass paramJC, string paramType)
        {
            VAMStorablesParam param = null;
            if (paramType == "Float") param = new VAMStorablesFloatParam(paramJC);
            if (paramType == "Bool") param = new VAMStorablesBoolParam(paramJC);
            if (paramType == "Color") param = new VAMStorablesColorParam(paramJC);
            if (paramType == "Vector3") param = new VAMStorablesVector3Param(paramJC);
            if (paramType == "String") param = new VAMStorablesStringParam(paramJC);
            if (paramType == "Url") param = new VAMStorablesUrlParam(paramJC);
            if (paramType == "StringChooser") param = new VAMStorablesStringChooserParam(paramJC, paramType);
            if (paramType == "StringChooserAction") param = new VAMStorablesStringChooserParam(paramJC, paramType);
            if (paramType == "Action") param = new VAMStorablesActionParam(paramJC, paramType);
            if (paramType == "PresetFilePathAction") param = new VAMStorablesActionParam(paramJC, paramType);
            if (paramType == "AudioClipAction") param = new VAMStorablesActionParam(paramJC, paramType);

            paramsByID.Add(param.paramID, param);
            if (!paramsByType.ContainsKey(paramType)) paramsByType.Add(paramType, new HashSet<VAMStorablesParam>());
            paramsByType[paramType].Add(param);
        }

        public void ProcessStorableVariations(JSONClass storableJC)
        {
            var paramTypesJC = storableJC["paramTypes"].AsObject;
            foreach (string paramType in paramTypes)
            {
                if (paramTypesJC[paramType] != null)
                {
                    foreach (JSONClass paramJC in paramTypesJC[paramType].AsArray)
                    {
                        string variationType = paramJC["variationType"];
                        string paramID = paramJC["paramID"];

                        if (paramsByID.ContainsKey(paramID) && variationType == "New") variationType = "Different";
                        if (variationType == "New") AddParam(paramJC, paramType);
                        if (variationType == "Different")
                        {
                            var param = paramsByID[paramID];
                            if (param.storableType ==JSONStorable.Type.StringChooser || param.storableType == JSONStorable.Type.StringChooserAction)
                            {
                                (param as VAMStorablesStringChooserParam).AddChoices(paramJC);
                            }
                        }
                    }
                }
            }
        }

        public List<string> GetParamNames()
        {
            return new List<string>(paramsByID.Keys);
        }

        public JSONStorable.Type GetJSONStorableType( string targetName)
        {
            return paramsByID[targetName].GetJSONStorableType();
        }

        public void ResetJSONParam(string targetName, JSONStorableParam param)
        {
            paramsByID[targetName].ResetJSONParam(param);
        }

    }

    public abstract class VAMStorablesParam
    {
        public JSONStorable.Type storableType { get; protected set; }
        public string paramID { get; protected set; }

        public JSONStorable.Type GetJSONStorableType()
        {
            return storableType;
        }

        public abstract void ResetJSONParam(JSONStorableParam param);

    }

    public class VAMStorablesFloatParam : VAMStorablesParam
    {
        private float min;
        private float max;
        private float defaultVal;
        private bool constrained;
        public VAMStorablesFloatParam(JSONClass paramJC)
        {
            paramID = paramJC["paramID"];
            storableType = JSONStorable.Type.Float;

            min = paramJC["min"].AsFloat;
            max = paramJC["max"].AsFloat;
            defaultVal = paramJC["defaultVal"].AsFloat;
            constrained = paramJC["constrained"].AsBool;
        }

        public JSONStorableFloat GetJSFParam(string jsfID)
        {
            return new JSONStorableFloat(jsfID, defaultVal, min, max, constrained);
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableFloat jsf = (JSONStorableFloat)param;
            jsf.min = min;
            jsf.max = max;
            jsf.defaultVal = defaultVal;
            jsf.val = defaultVal;

            if (min == 0 && max == 0) jsf.constrained = false;
            else jsf.constrained = constrained;
        }
    }

    public class VAMStorablesBoolParam : VAMStorablesParam
    {
        private bool defaultVal;

        public VAMStorablesBoolParam(JSONClass paramJC)
        {
            paramID = paramJC["paramID"];
            storableType = JSONStorable.Type.Bool;
            defaultVal = paramJC["defaultVal"].AsBool;
        }

        public JSONStorableBool GetJSBParam(string jsbID)
        {
            return new JSONStorableBool(jsbID, defaultVal);
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableBool jsb= (JSONStorableBool)param;
            jsb.defaultVal = defaultVal;
            jsb.val = defaultVal;
        }
    }

    public class VAMStorablesColorParam : VAMStorablesParam
    {
        private float defaultValH;
        private float defaultValS;
        private float defaultValV;

        public VAMStorablesColorParam(JSONClass paramJC)
        {
            paramID = paramJC["paramID"];
            storableType = JSONStorable.Type.Color;

            defaultValH = paramJC["defaultValH"].AsFloat;
            defaultValS = paramJC["defaultValS"].AsFloat;
            defaultValV = paramJC["defaultValV"].AsFloat;

        }

        public JSONStorableColor GetJSCParam(string jscID)
        {
            var hsvColor = new HSVColor();
            hsvColor.H= defaultValH;
            hsvColor.S= defaultValS;
            hsvColor.V= defaultValV;

            return new JSONStorableColor(jscID, hsvColor);
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableColor jsc = (JSONStorableColor)param;

            HSVColor defaultColor = new HSVColor();
            HSVColor currentColor = new HSVColor();

            defaultColor.H = defaultValH; defaultColor.S = defaultValS; defaultColor.V = defaultValV;
            currentColor.H = defaultValH; currentColor.S = defaultValS; currentColor.V = defaultValV;

            jsc.defaultVal = defaultColor;
            jsc.val = currentColor;
        }
    }

    public class VAMStorablesVector3Param : VAMStorablesParam
    {
        private float defaultValX;
        private float defaultValY;
        private float defaultValZ;

        private float minX;
        private float minY;
        private float minZ;

        private float maxX;
        private float maxY;
        private float maxZ;

        private bool constrained;

        public VAMStorablesVector3Param(JSONClass paramJC)
        {
            paramID = paramJC["paramID"];
            storableType = JSONStorable.Type.Vector3;

            defaultValX = paramJC["defaultValX"].AsFloat;
            defaultValY = paramJC["defaultValY"].AsFloat;
            defaultValZ = paramJC["defaultValZ"].AsFloat;

            minX = paramJC["minX"].AsFloat;
            minY = paramJC["minY"].AsFloat;
            minZ = paramJC["minZ"].AsFloat;

            maxX = paramJC["maxX"].AsFloat;
            maxY = paramJC["maxY"].AsFloat;
            maxZ = paramJC["maxZ"].AsFloat;

            constrained = paramJC["constrained"].AsBool;
        }

        public JSONStorableVector3 GetVector3Param(string jsv3ID)
        {
            return new JSONStorableVector3(jsv3ID,new Vector3(defaultValX,defaultValY,defaultValZ),new Vector3(minX, minY, minZ), new Vector3(maxX, maxY, maxZ),constrained);
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableVector3 jsv = param as JSONStorableVector3;

            Vector3 min = new Vector3(minX, minY, minZ);
            Vector3 max = new Vector3(maxX, maxY, maxZ);
            Vector3 def = new Vector3(defaultValX,defaultValY, defaultValZ);
            Vector3 val = new Vector3(defaultValX,defaultValY, defaultValZ);

            jsv.min = min; jsv.max = max;
            jsv.defaultVal= def;
            jsv.val = val;
        }
    }

    public class VAMStorablesStringParam : VAMStorablesParam
    {
        private string defaultVal;
        public VAMStorablesStringParam(JSONClass paramJC)
        {
            paramID = paramJC["paramID"];
            storableType = JSONStorable.Type.String;

            defaultVal = paramJC["defaultVal"];
        }

        public JSONStorableString GetStringParam(string jssID)
        {
            return new JSONStorableString(jssID, defaultVal);
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableString jss = param as JSONStorableString;
            jss.val= defaultVal;
            jss.defaultVal = defaultVal;
        }
    }

    public class VAMStorablesUrlParam : VAMStorablesParam
    {
        private string defaultVal;

        private bool showDirs;
        private bool allowFullComputerBrowse;
        private bool hideExtension;
        private bool allowBrowseAboveSuggestedPath;

        private string fileRemovePrefix;
        private string suggestedPath;
        private string suggestedPathGroup;
        public VAMStorablesUrlParam(JSONClass paramJC)
        {
            paramID = paramJC["paramID"];
            storableType = JSONStorable.Type.Url;
            defaultVal = paramJC["defaultVal"];

            showDirs = paramJC["showDirs"].AsBool;
            allowFullComputerBrowse = paramJC["allowFullComputerBrowse"].AsBool;
            hideExtension = paramJC["hideExtension"].AsBool;
            allowBrowseAboveSuggestedPath = paramJC["allowBrowseAboveSuggestedPath"].AsBool;

            fileRemovePrefix = paramJC["fileRemovePrefix"];
            suggestedPath = paramJC["suggestedPath"];
            suggestedPathGroup = paramJC["suggestedPathGroup"];
        }

        public JSONStorableUrl GetUrlParam(string jsUrlID)
        {
            var jsUrl = new JSONStorableUrl(jsUrlID, defaultVal);
            jsUrl.showDirs = showDirs;
            jsUrl.allowFullComputerBrowse= allowFullComputerBrowse;
            jsUrl.hideExtension = hideExtension;
            jsUrl.allowBrowseAboveSuggestedPath= allowBrowseAboveSuggestedPath;
            jsUrl.suggestedPath = suggestedPath;
            jsUrl.suggestedPathGroup = suggestedPathGroup;
            jsUrl.fileRemovePrefix= fileRemovePrefix;

            return jsUrl;
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableUrl jsu = param as JSONStorableUrl;
            jsu.val = defaultVal;
            jsu.defaultVal = defaultVal;
            jsu.showDirs = showDirs;
            jsu.allowFullComputerBrowse= allowFullComputerBrowse;
            jsu.hideExtension = hideExtension;
            jsu.allowBrowseAboveSuggestedPath = allowBrowseAboveSuggestedPath;
            jsu.suggestedPath = suggestedPath;
            jsu.fileRemovePrefix= fileRemovePrefix;
            jsu.suggestedPathGroup = suggestedPathGroup;
        }


    }

    public class VAMStorablesStringChooserParam : VAMStorablesParam
    {
        private string defaultVal;
        private List<string> choices;
        private List<string> displayChoices =null;

        public VAMStorablesStringChooserParam(JSONClass paramJC, string paramType)
        {
            paramID = paramJC["paramID"];
            if (paramType == "StringChooser") storableType = JSONStorable.Type.StringChooser;
            if (paramType == "StringChooserAction") storableType = JSONStorable.Type.StringChooserAction;
            defaultVal = paramJC["defaultVal"];

            choices = new List<string>();

            foreach (JSONNode choiceJN in paramJC["choices"].AsArray) choices.Add(choiceJN);

            if (paramJC["displayChoices"] != null)
            {
                displayChoices = new List<string>();
                foreach (JSONNode displayChoiceJN in paramJC["displayChoices"].AsArray) displayChoices.Add(displayChoiceJN);
            }
        }

        public void AddChoices (JSONClass paramJC)
        {
            if (paramJC["choices"] != null)
            {
                if (choices == null) choices = new List<string>();
                foreach (JSONNode choiceJN in paramJC["choices"].AsArray)
                {
                    if (!choices.Contains(choiceJN)) choices.Add(choiceJN);
                }
            }
                

            if (paramJC["displayChoices"] != null)
            {
                if (displayChoices == null) displayChoices = new List<string>();
                foreach (JSONNode displayChoiceJN in paramJC["displayChoices"].AsArray)
                {
                    if (!displayChoices.Contains(displayChoiceJN)) displayChoices.Add(displayChoiceJN);
                }
            }
        }

        public JSONStorableStringChooser GetStringChooserParam(string jssclID, string displayName)
        {
            List<string> choices = new List<string>(this.choices);
            List<string> displayChoices = null;

            if (this.displayChoices != null) displayChoices = new List<string>(this.displayChoices);

            return new JSONStorableStringChooser(jssclID, choices, displayChoices, defaultVal, displayName);
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            JSONStorableStringChooser jssc = (JSONStorableStringChooser)param;
            jssc.defaultVal = defaultVal;
            jssc.val= defaultVal;
            jssc.choices = new List<string>(choices);
            if (displayChoices != null) jssc.displayChoices = new List<string>(displayChoices);
            else jssc.displayChoices = jssc.choices;
            
        }

    }


    public class VAMStorablesActionParam : VAMStorablesParam
    {
        public VAMStorablesActionParam(JSONClass paramJC, string paramType)
        {
            paramID = paramJC["paramID"];
            if (paramType=="Action") storableType = JSONStorable.Type.Action;
            if (paramType == "PresetFilePathAction") storableType = JSONStorable.Type.PresetFilePathAction;
            if (paramType == "AudioClipAction") storableType = JSONStorable.Type.AudioClipAction;
        }

        public override void ResetJSONParam(JSONStorableParam param)
        {
            // Nothing to reset
        }
    }

}
